from mock import patch
from unittest import skipIf
import io
import sys

from future.moves.urllib.parse import parse_qs
from future.utils import PY3
from twisted.trial.unittest import SynchronousTestCase
from twisted.web.resource import IResource
from zope.interface.verify import verifyObject

from minion.core import Application
from minion.http import Headers
from minion.request import Response
from minion.tests.test_integration import RequestIntegrationTestMixin
from minion.twisted import MinionResource


if PY3:
    class EverythingIsTerrible(object):
        Klein = run = route = resource = flattenString = None
        Element = object
        XMLString = str
        TagLoader = object

        def renderer(fn):
            pass

    with patch.dict(
        sys.modules, {
            "StringIO": io,
            "klein.app": EverythingIsTerrible,
            "twisted.web.template": EverythingIsTerrible,
        },
    ):
        from klein.test_resource import (
            _render as render, requestMock as _requestMock,
        )
else:
    from klein.test_resource import (
            _render as render, requestMock as _requestMock,
        )


class TestMinionResource(SynchronousTestCase):
    def setUp(self):
        self.minion = Application()
        self.resource = MinionResource(self.minion)

    def assertRedirected(self, request, response, url, content=None):
        self.assertEqual(request.code, 302)
        self.assertEqual(
            request.responseHeaders.getRawHeaders("Location"), [url],
        )

        if content is not None:
            self.assertEqual(content, request.getWrittenData())

    @skipIf(PY3, "twisted.web doesn't support Py3 yet")
    def test_render(self):
        @self.minion.route(b"/foo/bar")
        def fooBar(request):
            self.assertEqual(request.headers.get(b"X-Foo"), [b"Hello"])
            return Response(
                code=302,
                content=b"/" + u"/".join(request.url.path).encode("ascii"),
                headers=Headers([(b"Location", [b"http://example.com"])]),
            )

        request = makeRequest(
            path=b"/foo/bar", headers={b"X-Foo": [b"Hello"]},
        )
        response = render(resource=self.resource, request=request)
        self.assertRedirected(
            request, response, b"http://example.com", content=b"/foo/bar",
        )

    @skipIf(PY3, "twisted.web doesn't support Py3 yet")
    def test_request_body(self):
        @self.minion.route(b"/")
        def respond(request):
            return Response(content=request.content.read())

        request = makeRequest(path=b"/", body=b"Hello world")
        render(resource=self.resource, request=request)
        self.assertEqual(request.getWrittenData(), b"Hello world")

    def test_interface(self):
        verifyObject(IResource, self.resource)


class TestRequestIntegration(RequestIntegrationTestMixin, SynchronousTestCase):
    def get(self, url, headers):
        request = makeRequest(path=url, headers=dict(headers.canonicalized()))
        render(resource=MinionResource(self.minion), request=request)
        return request.getWrittenData()

    def test_it_properly_routes_when_mounted_on_a_subpath(self):
        self.minion.route(b"/hello")(lambda request: Response(b"World"))
        request = makeRequest(path=b"/minion/hello")
        request.prepath.append(request.postpath.pop(0))
        render(resource=MinionResource(self.minion), request=request)
        self.assertEqual(request.getWrittenData(), b"World")


def makeRequest(path, *args, **kwargs):
    """
    Wrap Klein's request mock to support query strings and host headers.

    """

    path, _, queryString = path.partition(b"?")
    request = _requestMock(path=path, *args, **kwargs)
    request.args = parse_qs(queryString.rpartition(b"#")[0])

    # klein has a bug where it overrides the host when you call requestMock
    host = kwargs.get("headers", {}).get(b"Host", [b"localhost"])[0]
    request.setHost(host, kwargs.get("port", 80))

    return request



import attr

from minion import assets
from minion.request import Manager
from minion.routing import Router, SimpleMapper


@attr.s
class Application(object):
    """
    A Minion application.

    Arguments:

        config (collections.MutableMapping):

            any app configuration

        bin (assets.Bin):

            an asset bin containing assets used by views.

            If unprovided, an empty one will be created (until it is
            populated).

        manager (request.Manager):

            a request manager, which managers state during each request.

            If unprovided, one will be created and used.

        router:

            an object satisfying the router interface (see `minion.routing`) to
            use for route addition and generation for this application.

            If unprovided, a router with simple dictionary lookup will be used.

        jinja (jinja2.Environment):

            a pre-configured jinja2 environment, if using jinja2 is desired.

            (One can be added later, or multiple environments used, by calling
            `bind_jinja_environment`.)

    """

    config = attr.ib(default=attr.Factory(dict))
    manager = attr.ib(default=attr.Factory(Manager), repr=False)
    bin = attr.ib(default=None, repr=False)
    router = attr.ib(
        default=attr.Factory(lambda: Router(mapper=SimpleMapper())),
    )
    _jinja = attr.ib(default=None)

    def __attrs_post_init__(self):
        if self.bin is None:
            self.bin = assets.Bin().add(manager=self.manager)
        self.bin = self.bound_bin(self.bin)
        if self._jinja is not None:
            self.bind_jinja_environment(self._jinja)

    def route(self, route, **kwargs):
        def _add_route(fn):
            self.router.add(route, fn, **kwargs)
            return fn
        return _add_route

    def serve(self, request, path):
        self.manager.request_started(request)
        response = self.router.route(request=request, path=path)
        self.manager.request_served(request, response)
        return response

    def bound_bin(self, bin):
        """
        Bind an asset bin to this application.

        Returns:

            Bin: a bin containing this application's relevant globals

        """

        return bin.add(
            app=lambda bin: self,
            config=lambda bin: self.config,
            manager=lambda bin: self.manager,
            router=lambda bin: self.router,
        )

    def bind_jinja_environment(self, environment, asset_name="jinja"):
        """
        Bind useful pieces of the application to the given Jinja2 environment.

        Arguments:

            environment (jinja2.Environment):

                the environment to bind

            asset_name (str):

                a name to bind to in the application's asset bin.

                The default is ``'jinja'``.

        """

        environment.globals.update(
            [
                ("app", self),
                ("config", self.config),
                ("router", self.router),
            ],
        )
        self.bin = self.bin.add(**{asset_name: lambda bin: environment})



from __future__ import absolute_import

from hyperlink import URL
from twisted.web.resource import IResource
from zope.interface import implementer

from minion.http import Headers
from minion.request import Request


@implementer(IResource)
class MinionResource(object):
    """
    Wrap a Minion application in a :class:`twisted.web.resource.IResource`\ .

    Arguments:

        application (Application):

            a minion app
    """

    isLeaf = True

    def __init__(self, application):
        self.application = application

    def getChildWithDefault(self, path, request):
        return self

    def putChild(self, path, child):
        raise NotImplementedError()

    def render(self, twistedRequest):
        path = twistedRequest.uri.lstrip("/").decode("ascii").split(u"/")
        request = Request(
            content=twistedRequest.content,
            headers=Headers(twistedRequest.requestHeaders.getAllRawHeaders()),
            method=twistedRequest.method,
            url=URL(
                scheme=u"https" if twistedRequest.isSecure() else u"http",
                host=twistedRequest.getRequestHostname().decode("ascii"),
                path=path,
                query=[
                    (k.decode("ascii"), each.decode("ascii"))
                    for k, v in twistedRequest.args.items()
                    for each in v
                ],
            ),
        )
        response = self.application.serve(
            request=request, path=b"/" + b"/".join(twistedRequest.postpath),
        )

        twistedRequest.setResponseCode(response.code)

        for k, v in response.headers.canonicalized():
            twistedRequest.responseHeaders.setRawHeaders(k, v)

        return response.content



try:
    from collections.abc import MutableMapping
except ImportError:
    from collections import MutableMapping


class UnencryptedCookieSession(MutableMapping):
    def __init__(self):
        self._contents = {}

    def __iter__(self):
        return iter(self._contents)

    def __len__(self):
        return len(self._contents)

    def __delitem__(self, key):
        del self._contents[key]

    def __getitem__(self, key):
        return self._contents[key]

    def __setitem__(self, key, value):
        self._contents[key] = value



from unittest import TestCase, skip

from minion.deferred import Deferred


class ABadThing(Exception):
    pass


class TestDeferred(TestCase):

    success = error = second = None

    def assertResultedIn(self, success=None, error=None, second=None):
        self.assertEqual(
            {"success": self.success, "error": self.error},
            {"success": success, "error": error},
        )
        self.assertEqual(self.second, second)

    def _succeeded(self, arg, *rest, **kw):
        self.success = (arg,) + rest, kw
        return arg

    def _second_succeeded(self, *args, **kw):
        self.second = args, kw

    def _errored(self, *args, **kw):
        self.error = args, kw

    def test_succeed(self):
        deferred = Deferred()
        deferred.chain.on_success(self._succeeded, "world", end="!")
        deferred.succeed("hello")
        self.assertResultedIn(success=(("hello", "world"), {"end": "!"}))

    def test_succeed_without_args(self):
        deferred = Deferred()
        deferred.chain.on_success(self._succeeded)
        deferred.succeed("hello")
        self.assertResultedIn(success=(("hello",), {}))

    def test_two_on_successes(self):
        deferred = Deferred()
        deferred.chain.on_success(self._succeeded)
        deferred.chain.on_success(self._second_succeeded)
        deferred.succeed("hello")
        result = (("hello",), {})
        self.assertResultedIn(success=result, second=result)

    def test_immediate_on_success(self):
        deferred = Deferred()
        deferred.succeed("hello")
        deferred.chain.on_success(self._succeeded)
        self.assertResultedIn(success=(("hello",), {}))

    @skip("No errback support yet")
    def test_error(self):
        deferred = Deferred()
        deferred.chain.on_error(self._errored)
        error = ABadThing("Whoops!")
        deferred.error(error)
        self.assertResultedIn(error=((error,), {}))



from functools import partial, wraps
import json

import attr

from minion.http import Headers, MediaRange
from minion.request import Response


class JSON(object):
    """
    An 'intelligent' JSON renderer that pretty-prints JSON for humans.

    """

    def __init__(self, **kwargs):
        self._dumps = partial(json.dumps, **kwargs)

    def render(self, request, jsonable):
        machine_json = request.accept.media_types[-1] == MediaRange(
            type="application", subtype="json",
        )
        if machine_json:
            content = self._dumps(jsonable, separators=",:")
        else:
            content = self._dumps(jsonable, indent=2, sort_keys=True)
        return Response(
            content=content,
            headers=Headers([("Content-Type", ["application/json"])]),
        )


class SimpleJSON(object):
    """
    A simple JSON renderer that renders by dumping with any given parameters.

    """

    def __init__(self, **kwargs):
        self._dumps = partial(json.dumps, **kwargs)

    def render(self, request, jsonable):
        return Response(self._dumps(jsonable))


@attr.s
class Unicode(object):

    encoding = attr.ib()
    errors = attr.ib(default="strict")

    def render(self, request, text):
        return Response(
            text.encode(encoding=self.encoding, errors=self.errors),
        )


UTF8 = Unicode(encoding="utf-8")


def bind(renderer, to):
    """
    Bind a renderer to the given callable by constructing a new rendering view.

    """

    @wraps(to)
    def view(request, **kwargs):
        try:
            returned = to(request, **kwargs)
        except Exception as error:
            view_error = getattr(renderer, "view_error", None)
            if view_error is None:
                raise
            return view_error(request, error)

        try:
            return renderer.render(request, returned)
        except Exception as error:
            render_error = getattr(renderer, "render_error", None)
            if render_error is None:
                raise
            return render_error(request, returned, error)

    return view



from pyrsistent import m
import attr


class DuplicateAsset(Exception):
    pass


class NoSuchAsset(LookupError):
    pass


@attr.s(hash=True, repr=False)
class Bin(object):

    _assets = attr.ib(default=m())

    def __repr__(self):
        return "<{}: {{{}}}>".format(
            self.__class__.__name__,
            ", ".join(repr(asset) for asset in sorted(self._assets)),
        )

    def add(self, **assets):
        return attr.evolve(self, assets=self._assets.update(assets))

    def with_globals(self, **kwargs):
        def _global(thing):
            return lambda bin: thing
        return self.add(
            **{name: _global(thing) for name, thing in kwargs.items()}
        )

    def provide(self, name):
        try:
            provider = self._assets[name]
        except KeyError:
            raise NoSuchAsset(name)
        else:
            return provider(bin=self)



from unittest import TestCase

from hyperlink import URL

from minion import traversal
from minion.request import Request, Response


def path_view(request):
    """
    A renderer (view) that returns a response with the request path.

    Optionally adds a prefix to the front.

    """

    path = b"/" + b"/".join(each.encode("ascii") for each in request.url.path)
    return Response(content=path)


class TestTreeResource(TestCase):
    def test_it_renders_what_its_told(self):
        resource = traversal.TreeResource(render=path_view)
        request = Request(url=URL(path=[u"bar"]))
        self.assertEqual(resource.render(request), Response(content=b"/bar"))

    def test_it_supports_adding_children(self):
        resource = traversal.TreeResource(render=path_view)
        child = traversal.LeafResource(render=path_view)
        resource.set_child(b"foo", child)
        request = Request(url=URL(path=[u"bar"]))
        self.assertEqual(resource.get_child(b"foo", request=request), child)

    def test_it_returns_404s_for_unknown_children(self):
        resource = traversal.TreeResource(render=path_view)
        request = Request(url=URL(path=[u"foo"]))
        child = resource.get_child(b"foo", request=request)
        self.assertEqual(child.render(request), Response(code=404))


class TestLeafResource(TestCase):
    def test_it_renders_what_its_told(self):
        resource = traversal.LeafResource(render=path_view)
        request = Request(url=URL(path=[u"foo"]))
        self.assertEqual(resource.render(request), Response(content=b"/foo"))

    def test_it_is_a_leaf_resource(self):
        self.assertTrue(
            traversal.LeafResource(render=lambda request: Response()).is_leaf,
        )


class TestMethodDelegate(TestCase):
    def test_known_HTTP_methods_use_associated_renderer(self):
        render = traversal.method_delegate(
            GET=lambda request: b"foo",
            put=lambda request: request.method,
        )
        get = Request(url=URL(path=[u""]), method=b"GET")
        self.assertEqual(render(get), b"foo")

        put = Request(url=URL(path=[u""]), method=b"PUT")
        self.assertEqual(render(put), b"PUT")

    def test_unknown_HTTP_methods_return_405s(self):
        render = traversal.method_delegate(get=lambda _: b"foo")
        request = Request(url=URL(path=[u""]), method=b"PUT")
        self.assertEqual(render(request).code, 405)


class LineDelimiterResource(object):
    """
    A resource where traversal returns a child resource that renders the
    ultimate path as a line delimited bytestring.

    """

    def __init__(self, path=()):
        self.path = list(path)

    def get_child(self, name, request):
        return self.__class__(path=self.path + [name])

    def render(self, request):
        return b"\n".join(self.path)


class TestTraverse(TestCase):
    def test_it_traverses_resources(self):
        root = LineDelimiterResource()
        request = Request(url=URL(path=[u"foo", u"bar", u"baz"]))
        renderer = traversal.traverse(
            path=b"/foo/bar/baz", resource=root, request=request,
        )
        self.assertEqual(renderer.render(request), b"foo\nbar\nbaz")

    def test_single_level(self):
        root = LineDelimiterResource()
        request = Request(url=URL(path=[u"foo"]))
        renderer = traversal.traverse(
            path=b"/foo", resource=root, request=request,
        )
        self.assertEqual(renderer.render(request), b"foo")

    def test_zero_levels(self):
        root = LineDelimiterResource()
        request = Request(url=URL(path=[u""]))
        renderer = traversal.traverse(
            path=b"/", resource=root, request=request,
        )
        self.assertIs(renderer, root)

    def test_leaf_resources_are_not_traversed(self):
        class Resource(object):
            def __init__(self, name=b""):
                self.name = name

            def get_child(self, name, request):
                resource = Resource(name=name)
                if name == b"2":
                    resource.is_leaf = True
                return resource

            def render(self, name):
                return self.name

        root = Resource()
        request = Request(url=URL(path=[u"0", u"1", u"2", u"3", u"4", u"5"]))
        renderer = traversal.traverse(
            path=b"/0/1/2/3/4/5", resource=root, request=request,
        )
        self.assertEqual(renderer.render(request), b"2")



"""
flaskr -- the Flask tutorial app. Or, should I say, Minion.io

"""

from contextlib import closing
import os
import sqlite3

import jinja2
import werkzeug.wrappers

from minion.core import Application
from minion.request import Response
from minion.routing import Router, WerkzeugMapper
from minion.wsgi import create_app
import examples.static


CSS = os.path.join(os.path.dirname(examples.static.__file__), "flaskr.css")
loader = jinja2.FileSystemLoader(
    os.path.join(os.path.dirname(__file__), "templates", "flaskr"),
)
app = Application(
    jinja=jinja2.Environment(loader=loader),
    router=Router(mapper=WerkzeugMapper()),
    config = {
        "credentials" : {"user" : "admin", "password" : "default"},
        "database" : {"uri" : "/tmp/flaskr.db"},
    },
)


@app.route("/")
@app.bin.needs(["db", "jinja"])
def show_entries(request, db, jinja):
    cur = db.execute("SELECT title, text FROM entries ORDER BY id DESC")
    entries = cur.fetchall()
    content = jinja.get_template("show_entries.html").render(entries=entries)
    return Response(content.encode("utf-8"))


@app.route('/add', methods=['POST'])
@app.bin.needs(["db", "jinja"])
def add_entry(request, db, jinja):
    # if not session.get('logged_in'):
    #     return Response(code=401)
    db.execute(
        "INSERT INTO entries (title, text) VALUES (?, ?)",
        [request.form["title"], request.form["text"]]
    )
    db.commit()
    # flash('New entry was successfully posted')
    # return redirect(url_for('show_entries'))
    return Response(code=302, headers={"Location" : "/"})


@app.route("/login", methods=["GET", "POST"])
@app.bin.needs(["config", "jinja"])
def login(request, config, jinja):
    error = None
    if request.method == "POST":
        if request.form["username"] != config["credentials"]["user"]:
            error = "Invalid username"
        elif request.form["password"] != config["credentials"]["password"]:
            error = "Invalid password"
        else:
            # session["logged_in"] = True
            # flash("You were logged in")
            # return redirect(url_for("show_entries"))
            return Response(code=302, headers={"Location" : "/"})
    content = jinja.get_template("login.html").render(error=error)
    return Response(content.encode("utf-8"))


@app.route("/logout")
def logout(request):
    # session.pop("logged_in", None)
    # flash("You were logged out")
    # return redirect(url_for("show_entries"))
    return Response(code=302, headers={"Location" : "/"})


@app.route("/static/flaskr.css")
def style(request):
    with open(CSS) as css:
        return Response(css.read())


@app.bin.provides("db")
@app.bin.needs(["config"])
def connect_db(config):
    """Connects to the specific database."""
    rv = sqlite3.connect(config["database"]["uri"])
    rv.row_factory = sqlite3.Row
    return rv


def init_db():
    """Creates the database tables."""
    with closing(connect_db()) as db:
        db.cursor().execute("DROP TABLE IF EXISTS entries")
        db.cursor().execute(
            """
            CREATE TABLE entries (
                id INTEGER PRIMARY KEY AUTOINCREMENT,
                title TEXT NOT NULL,
                text TEXT NOT NULL
            )
            """
        )
        db.commit()


wsgi = create_app(app)


if __name__ == "__main__":
    init_db()



from unittest import TestCase
import mock

from hyperlink import URL

from minion import request
from minion.http import Accept, Headers
from minion.tests.utils import ResponseHelpersMixin


class TestResponder(TestCase):
    def setUp(self):
        self.request = mock.Mock()
        self.responder = request.Responder(request=self.request)

    def test_it_runs_deferreds_registered_with_after(self):
        m = mock.Mock()
        self.responder.after().on_success(m)
        self.assertFalse(m.called)

        self.responder.finish()

        m.assert_called_once_with(self.responder)


class TestManager(TestCase):
    def setUp(self):
        self.manager = request.Manager()
        self.request = mock.Mock()
        self.response = mock.Mock()

    def test_after_response(self):
        self.manager.request_started(self.request)
        callback = mock.Mock(return_value=None)

        self.manager.after_response(self.request, callback, 1, kw="abc")

        response = self.manager.request_served(self.request, self.response)

        self.assertEqual(response, self.response)
        callback.assert_called_once_with(self.response, 1, kw="abc")

    def test_after_response_can_modify_the_response(self):
        new_response = mock.Mock()

        self.manager.request_started(self.request)

        def set_thing(response):
            return new_response

        self.manager.after_response(self.request, set_thing)

        response = self.manager.request_served(self.request, self.response)

        self.assertEqual(response, new_response)

    def test_after_response_chaining(self):
        self.manager.request_started(self.request)

        def set_thing(response):
            response.thing = 2

        def add_2(response):
            response.thing += 2

        self.manager.after_response(self.request, set_thing)
        self.manager.after_response(self.request, add_2)

        response = self.manager.request_served(self.request, self.response)

        self.assertEqual(response, self.response)
        self.assertEqual(response.thing, 4)


class RequestTestMixin(object):
    """
    Tests implementations of the Request interface.

    """

    def test_accept(self):
        accept = "application/json"
        request = self.make_request(headers=Headers([("accept", [accept])]))
        self.assertEqual(request.accept, Accept.from_header(accept))

    def test_accept_no_header(self):
        request = self.make_request(headers=Headers())
        self.assertEqual(request.accept, Accept.ALL)

    def test_accept_is_calculated_once(self):
        accept = "application/json"
        request = self.make_request(headers=Headers([("accept", [accept])]))
        self.assertIs(request.accept, request.accept)


class TestRequest(RequestTestMixin, TestCase):
    def make_request(self, headers):
        return request.Request(url=URL(path=[u""]), headers=headers)

    def test_url(self):
        self.request = request.Request(url=URL(path=[u"foo", u"bar"]))
        self.assertEqual(self.request.url, URL(path=[u"foo", u"bar"]))

    def test_flash(self):
        self.request = request.Request(url=URL(path=[u""]))
        self.assertEqual(self.request.messages, [])
        self.request.flash(b"Hello World")
        self.assertEqual(
            self.request.messages, [request._Message(content=b"Hello World")],
        )


class TestResponse(TestCase):
    def test_init(self):
        self.assertEqual(request.Response(b"Hello").content, b"Hello")

    def test_init_kwargs(self):
        self.assertEqual(request.Response(content=b"Hello").content, b"Hello")


class TestRedirect(ResponseHelpersMixin, TestCase):
    def test_it_returns_redirect_responses(self):
        response = request.redirect(to="http://example.com")
        self.assertRedirected(response, "http://example.com")

    def test_it_allows_specifying_a_status_code(self):
        response = request.redirect(to="http://example.com", code=304)
        self.assertRedirected(response, "http://example.com", code=304)



"""
Helpers and utilities for testing Minion applications.

Unlike the rest of the test modules, this module *is* public API.
"""


class ResponseHelpersMixin(object):
    def assertRedirected(self, response, to, code=302):
        """
        Assert that the given response is a redirect to the given URL.

        """

        self.assertEqual(response.code, code)
        self.assertIn("Location", response.headers)
        self.assertEqual(response.headers.get("Location")[0], to)



from unittest import TestCase, skipIf

from future.utils import PY3
from hyperlink import URL
from webtest import TestApp
from werkzeug.test import create_environ

from minion import wsgi
from minion.core import Application
from minion.http import Headers
from minion.request import Response
from minion.tests.test_integration import RequestIntegrationTestMixin
from minion.tests.test_request import RequestTestMixin


@skipIf(PY3, "WSGI is pure insanity on Py3")
class TestWSGIMinion(TestCase):
    def setUp(self):
        self.minion = Application()
        self.wsgi = TestApp(wsgi.create_app(self.minion))

    def test_it_speaks_wsgi(self):
        @self.minion.route(b"/respond")
        def respond(request):
            return Response(b"Yep!")

        response = self.wsgi.get("/respond", status=200)
        self.assertEqual(response.body, b"Yep!")

    def test_it_parses_headers(self):
        @self.minion.route(b"/respond")
        def respond(request):
            return Response(request.headers.get("Accept")[0])

        response = self.wsgi.get(
            b"/respond", status=200, headers={b"Accept": b"2"},
        )
        self.assertEqual(response.body, b"2")

    def test_it_parses_response_bodies(self):
        @self.minion.route(b"/respond", methods=["POST"])
        def respond(request):
            return Response(request.content.read())

        response = self.wsgi.post(
            b"/respond",
            params=b"Check out this body.",
            status=200,
        )
        self.assertEqual(response.body, b"Check out this body.")

    def test_it_sets_headers(self):
        @self.minion.route(b"/respond")
        def respond(request):
            return Response(
                b"{}",
                headers=Headers([(b"Content-Type", [b"application/json"])]),
            )

        response = self.wsgi.get(b"/respond", status=200)
        self.assertEqual(
            response.headers[b"Content-Type"],
            b"application/json",
        )


class TestRequest(RequestTestMixin, TestCase):
    def make_request(self, headers):
        headers = {k: b",".join(v) for k, v in headers.canonicalized()}
        return wsgi.Request(environ=create_environ(headers=headers))

    def test_query_string(self):
        environ = {
            "SERVER_NAME": "example.com",
            "SERVER_PORT": "80",
            "QUERY_STRING": "foo=bar",
            "wsgi.url_scheme": "http",
        }
        self.assertEqual(
            wsgi.Request(environ).url, URL(
                scheme=u"http",
                host=u"example.com",
                query=[(u"foo", u"bar")],
            ),
        )

    def test_default_port(self):
        environ = {
            "SERVER_NAME": "localhost",
            "SERVER_PORT": "80",
            "wsgi.url_scheme": "http",
        }
        self.assertEqual(
            wsgi.Request(environ).url,
            URL(scheme=u"http", host=u"localhost", port=None),
        )

    def test_non_default_port(self):
        environ = {
            "SERVER_NAME": "localhost",
            "SERVER_PORT": "8080",
            "wsgi.url_scheme": "http",
        }
        self.assertEqual(
            wsgi.Request(environ).url,
            URL(scheme=u"http", host=u"localhost", port=8080),
        )

    def test_http_host_present(self):
        environ = {
            "HTTP_HOST": "example.org:1234",
            "PATH_INFO": "/",
            "SERVER_PORT": "1234",
            "wsgi.url_scheme": "https",
        }
        self.assertEqual(
            wsgi.Request(environ).url, URL(
                scheme=u"https",
                host=u"example.org",
                path=[u""],
                port=1234,
            ),
        )

    def test_http_host_present_default_port(self):
        environ = {
            "HTTP_HOST": "example.org:80",
            "SERVER_PORT": "80",
            "PATH_INFO": "/",
            "wsgi.url_scheme": "https",
        }
        self.assertEqual(
            wsgi.Request(environ).url, URL(
                scheme=u"https",
                host=u"example.org",
                path=[u""],
                port=80,
            ),
        )

    def test_no_http_host_present(self):
        environ = {
            "SERVER_NAME": "example.org",
            "SERVER_PORT": "1234",
            "PATH_INFO": "/",
            "wsgi.url_scheme": "https",
        }
        self.assertEqual(
            wsgi.Request(environ).url, URL(
                scheme=u"https",
                host=u"example.org",
                path=[u""],
                port=1234,
            ),
        )

    def test_script_name(self):
        environ = {
            "HTTP_HOST": "example.org",
            "SCRIPT_NAME": "/stuff",
            "PATH_INFO": "/things",
            "SERVER_PORT": "443",
            "wsgi.url_scheme": "https",
        }
        self.assertEqual(
            wsgi.Request(environ).url, URL(
                scheme=u"https",
                host=u"example.org",
                path=[u"stuff", u"things"],
            ),
        )

    def test_host_no_path_no_query_string(self):
        environ = {
            "HTTP_HOST": "example.com",
            "SERVER_NAME": "localhost",
            "SERVER_PORT": "80",
            "wsgi.url_scheme": "http",
        }
        self.assertEqual(
            wsgi.Request(environ).url,
            URL(scheme=u"http", host=u"example.com"),
        )

    def test_no_host_no_path_no_query_string(self):
        environ = {
            "SERVER_NAME": "localhost",
            "SERVER_PORT": "80",
            "wsgi.url_scheme": "http",
        }
        self.assertEqual(
            wsgi.Request(environ).url,
            URL(scheme=u"http", host=u"localhost"),
        )


class TestRequestIntegration(RequestIntegrationTestMixin, TestCase):
    def get(self, url, headers):
        app = TestApp(wsgi.create_app(self.minion))
        response = app.get(
            url,
            headers=[(k, b",".join(v)) for k, v in headers.canonicalized()],
        )
        return response.body

    def test_script_name(self):
        self.minion.route(b"/home")(lambda request: Response(b"Hi!"))
        app = TestApp(
            wsgi.create_app(self.minion),
            extra_environ={"SCRIPT_NAME": "/app"},
        )
        response = app.get(b"/app/home")
        self.assertEqual(response.body, b"Hi!")



from unittest import TestCase

from minion import sessions


class TestUnencryptedCookieSession(TestCase):
    def setUp(self):
        self.session = sessions.UnencryptedCookieSession()

    def test_its_a_mutable_mapping(self):
        self.session["thing"] = 12
        self.assertIn("thing", self.session)
        self.assertNotIn("other", self.session)
        self.assertEqual(self.session["thing"], 12)



from minion.core import Application
from minion.request import Response
from minion.wsgi import create_app


app = Application()


@app.route("/")
def index(request):
    return Response("Hello World!")


wsgi = create_app(app)



from unittest import TestCase

from minion import assets


class TestBin(TestCase):
    def test_it_provides_assets(self):
        bin = assets.Bin().add(twelve=lambda bin: 12)
        self.assertEqual(bin.provide("twelve"), 12)

    def test_with_globals(self):
        foo, bar = object(), object()
        bin = assets.Bin().with_globals(foo=foo, bar=bar)
        self.assertEqual((bin.provide("foo"), bin.provide("bar")), (foo, bar))

    def test_no_such_asset(self):
        with self.assertRaises(assets.NoSuchAsset) as e:
            assets.Bin().provide("twelve")
        self.assertEqual(str(e.exception), "twelve")

    def test_hash(self):
        self.assertIn(
            assets.Bin().add(zero=int),
            {assets.Bin().add(zero=int)},
        )

    def test_repr(self):
        self.assertEqual(
            repr(assets.Bin().add(zero=int).add(empty=str)),
            "<Bin: {'empty', 'zero'}>",
        )



from unittest import TestCase, skipIf

from future.utils import iteritems
from hyperlink import URL

try:
    import jinja2
except ImportError:
    jinja2 = None

from minion import core, assets
from minion.request import Manager, Request, Response
from minion.routing import Router, SimpleMapper


class TestApplication(TestCase):
    def assertSubdict(self, subdict, of):
        """
        Assert that the given thing is a subdict of the given dict.

        """

        subdict = dict(subdict)
        matches = {k: v for k, v in iteritems(of) if k in subdict}
        self.assertEqual(subdict, matches)

    def test_it_delegates_routes_to_the_router(self):
        class Router(object):
            def add(this, route, fn, baz):
                self.added = route, fn, baz

        application = core.Application(router=Router())

        @application.route("/foo", baz=2)
        def fn(request):
            pass

        self.assertEqual(getattr(self, "added", None), ("/foo", fn, 2))

    def test_binds_to_its_bin(self):
        app = core.Application()
        self.assertEqual(
            (
                app.bin.provide("app"),
                app.bin.provide("config"),
                app.bin.provide("manager"),
                app.bin.provide("router"),
            ), (
                app,
                app.config,
                app.manager,
                app.router,
            ),
        )

    def test_it_can_bind_to_other_bins(self):
        app = core.Application()
        bin = app.bound_bin(assets.Bin())
        self.assertEqual(
            (
                bin.provide("app"),
                bin.provide("config"),
                bin.provide("manager"),
                bin.provide("router"),
            ), (
                app,
                app.config,
                app.manager,
                app.router,
            ),
        )

    @skipIf(jinja2 is None, "jinja2 not found")
    def test_it_can_bind_to_jinja_environments(self):
        environment = jinja2.Environment()
        app = core.Application(jinja=environment)
        self.assertEqual(app.bin.provide("jinja"), environment)


class TestApplicationIntegration(TestCase):
    def setUp(self):
        self.manager = Manager()
        self.router = Router(mapper=SimpleMapper())
        self.app = core.Application(manager=self.manager, router=self.router)
        self.request = Request(url=URL(path=[u""]))

    def test_it_serves_mapped_requests(self):
        self.router.add(
            self.request.url.path, lambda request: Response(b"Hello"),
        )
        self.assertEqual(
            self.app.serve(self.request, path=self.request.url.path),
            Response(b"Hello"),
        )

    def test_it_serves_404s_for_unmapped_requests_by_default(self):
        self.assertEqual(
            self.app.serve(self.request, path=self.request.url.path).code, 404,
        )



"""
APIs for storing and retrieving HTTP headers and cookies.

"""

from bisect import insort

from future.utils import iteritems, viewkeys
from pyrsistent import m
import attr


_CANONICAL_HEADER_NAMES = {
    b"content-md5": b"Content-MD5",
    b"dnt": b"DNT",
    b"etag": b"ETag",
    b"p3p": b"P3P",
    b"te": b"TE",
    b"www-authenticate": b"WWW-Authenticate",
    b"x-xss-protection": b"X-XSS-Protection",
}
DEFAULT_PORTS = {b"http": 80, b"https": 443}


class NoSuchHeader(LookupError):
    """
    An operation on a non-present or -existing HTTP header was performed.

    """


class Headers(object):
    def __init__(self, contents=()):
        self._contents = dict(
            (name.lower(), values) for name, values in contents
        )

    def __contains__(self, name):
        return name.lower() in self._contents

    def __getitem__(self, name):
        try:
            return self._contents[name.lower()]
        except KeyError:
            raise NoSuchHeader(name)

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return self._contents == other._contents

    def __ne__(self, other):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return not self == other

    def __hash__(self):
        calculated = getattr(self, "_hash", None)
        if calculated is None:
            calculated = 0
            for name, values in iteritems(self._contents):
                calculated ^= hash((name, tuple(values)))
            self._hash = calculated
        return calculated

    def __repr__(self):
        contents = " ".join(
            "{0}={1!r}".format(name, values)
            for name, values in sorted(self.canonicalized())
        )
        return "<{.__class__.__name__} contents={{{}}}>".format(self, contents)

    def canonicalized(self):
        for name, values in iteritems(self._contents):
            canonical_name = _CANONICAL_HEADER_NAMES.get(name)
            if canonical_name is None:
                canonical_name = b"-".join(
                    word.capitalize() for word in name.split(b"-")
                )
            yield canonical_name, values

    def get(self, name, default=None):
        return self._contents.get(name.lower(), default)


class MutableHeaders(Headers):

    __hash__ = None

    def __setitem__(self, name, values):
        self._contents[name.lower()] = values

    def add_value(self, name, value):
        existing = self.get(name)
        if existing is None:
            self[name] = [value]
        else:
            existing.append(value)

    def get(self, name, default=None):
        return self._contents.get(name.lower(), default)

    def pop(self, name, *args, **kwargs):
        try:
            return self._contents.pop(name.lower(), *args, **kwargs)
        except KeyError:
            raise NoSuchHeader(name)


@attr.s
class Accept(object):
    """
    A parsed representation of an HTTP Accept header (see :rfc:`7231#5.3.2`\ ).

    """

    media_types = attr.ib()

    @classmethod
    def from_header(cls, header):
        """
        Parse out an Accept header.

        """

        if header is None:
            return cls.ALL

        media_types = []
        for range_and_parameters in header.split(b","):
            raw_range, _, raw_parameters = range_and_parameters.partition(b";")

            quality = 1.0
            media_parameters = {}
            if raw_parameters:
                for raw_parameter in raw_parameters.split(b";"):
                    key, _, value = raw_parameter.partition(b"=")
                    key = key.strip()
                    value = value.strip()
                    if key == b"q":
                        quality = float(value)
                    else:
                        media_parameters[key] = value

            raw_type, _, raw_subtype = raw_range.partition(b"/")

            type = raw_type.strip()
            subtype = raw_subtype.strip()

            insort(
                media_types,
                MediaRange(
                    type=type if type != b"*" else STAR,
                    subtype=subtype if subtype != b"*" else STAR,
                    quality=quality,
                    parameters=media_parameters,
                ),
            )
        return cls(media_types=tuple(media_types))


class _Star(object):
    """
    An open media range value.

    """

    def __lt__(self, other):
        return True

    def __repr__(self):
        return "*"


STAR = _Star()


@attr.s(cmp=False, hash=True)
class MediaRange(object):
    """
    A media range.

    Open ranges (e.g. ``text/*`` or ``*/*``\ ) are represented by
    :attr:`type` and / or :attr:`subtype` being ``None``\ .

    """

    type = attr.ib(default=STAR)
    subtype = attr.ib(default=STAR)
    parameters = attr.ib(default=m())
    quality = attr.ib(default=1.0)

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return attr.astuple(self) == attr.astuple(other)

    def __ne__(self, other):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return not self == other

    def __lt__(self, other):
        if not isinstance(other, self.__class__):
            return NotImplemented

        if self.quality < other.quality:
            return True
        elif self.quality > other.quality:
            return False

        if self.type != other.type:
            return self.type is STAR

        if self.subtype != other.subtype:
            return self.subtype is STAR

        return viewkeys(self.parameters) < viewkeys(other.parameters)


Accept.ALL = Accept(media_types=(MediaRange(),))



from minion.core import Application
from minion.request import Response
from minion.routing import Router, RoutesMapper
from minion.wsgi import create_app


app = Application(router=Router(mapper=RoutesMapper()))


@app.route("/greet/{user}", greeting="Hey")
def greet(request, greeting, user):
    return Response(u"{} {}\n".format(greeting, user).encode("utf-8"))


wsgi = create_app(app)



from cached_property import cached_property as calculated_once
from future.moves.urllib.parse import parse_qsl
from future.utils import iteritems
from hyperlink import URL

from minion.http import Accept, Headers


class Request(object):
    def __init__(self, environ):
        self.environ = environ

    @calculated_once
    def accept(self):
        return Accept.from_header(header=self.environ.get("HTTP_ACCEPT"))

    @calculated_once
    def headers(self):
        return Headers(
            (name[5:].replace("_", "-"), [value])
            for name, value in iteritems(self.environ)
            if name.startswith("HTTP_")
        )

    @calculated_once
    def content(self):
        return self.environ["wsgi.input"]

    @calculated_once
    def method(self):
        return self.environ["REQUEST_METHOD"]

    @calculated_once
    def url(self):
        environ = self.environ

        # See https://www.python.org/dev/peps/pep-0333/#url-reconstruction
        host = environ.get("HTTP_HOST")
        if host is not None:
            host, _, port = host.partition(":")
            port = int(port) if port else None
        else:
            host = environ["SERVER_NAME"]
            port = int(environ["SERVER_PORT"])

        path = environ.get("SCRIPT_NAME", "") + environ.get("PATH_INFO", "")
        return URL(
            host=host.decode("ascii"),
            port=port,
            path=path.lstrip("/").decode("ascii").split(u"/"),
            query=[
                (k.decode("ascii"), v.decode("ascii"))
                for k, v in parse_qsl(environ.get("QUERY_STRING", ""))
            ],
            scheme=environ["wsgi.url_scheme"].decode("ascii"),
        )


def create_app(application, request_class=Request):
    """
    Create a WSGI application out of the given Minion app.

    Arguments:

        application (Application):

            a minion app

        request_class (callable):

            a class to use for constructing incoming requests out of the WSGI
            environment. It will be passed a single arg, the environ.

            By default, this is :class:`minion.request.WSGIRequest` if
            unprovided.
    """

    def wsgi(environ, start_response):
        response = application.serve(
            request=request_class(environ),
            path=environ.get("PATH_INFO", ""),
        )
        start_response(
            response.status, [
                (name, b",".join(values))
                for name, values in response.headers.canonicalized()
            ],
        )
        return [response.content]
    return wsgi



from unittest import TestCase

from future.utils import PY3
from pyrsistent import pmap

from minion import http


class HeaderRetrievalTestsMixin(object):
    """
    Test that headers can be retrieved from a populated Header instance.

    """

    def test_init(self):
        headers = self.Headers([(b"Foo", [b"Bar"])])
        self.assertEqual(headers.get(b"Foo"), [b"Bar"])

    def test_contains(self):
        headers = self.Headers([(b"test", [b"hello", b"world"])])
        self.assertIn(b"test", headers)

    def test_does_not_contain(self):
        headers = self.Headers()
        self.assertNotIn(b"test", headers)

    def test_getitem(self):
        headers = self.Headers([(b"thing", [b"value"])])
        self.assertEqual(headers[b"tHiNg"], [b"value"])
        self.assertIn(b"thing", headers)

    def test_getitem_missing(self):
        with self.assertRaises(http.NoSuchHeader):
            self.Headers()[b"thiNg"]

    def test_get(self):
        headers = self.Headers([(b"thiNg", [b"value"])])
        self.assertEqual(headers.get(b"thing"), [b"value"])
        self.assertIn(b"thing", headers)

    def test_get_with_default_returning_default(self):
        headers = self.Headers()
        default = object()
        self.assertIs(headers.get(b"nonexisting", default), default)

    def test_get_with_default_returning_value(self):
        headers = self.Headers([(b"existing", [b"value"])])
        default = object()
        self.assertEqual(headers.get(b"existing", default), [b"value"])
        self.assertIn(b"existing", headers)

    def test_get_defaults_to_None(self):
        headers = self.Headers()
        self.assertIsNone(headers.get(b"nonexisting"))

    def test_canonicalized(self):
        headers = self.Headers(
            [
                (b"red", [b"Blue"]),
                (b"yellow-and-green", [b"purple-too"]),
                (b"content-md5", [b"Stuff"]),
                (b"dnT", [b"0"]),
                (b"etAg", [b"more"]),
                (b"p3p", [b"stuff"]),
                (b"te", [b"things"]),
                (b"www-authenticate", [b"hooray"]),
                (b"x-xss-protection", [b"done"]),
            ],
        )

        self.assertEqual(
            sorted(headers.canonicalized()), [
                (b"Content-MD5", [b"Stuff"]),
                (b"DNT", [b"0"]),
                (b"ETag", [b"more"]),
                (b"P3P", [b"stuff"]),
                (b"Red", [b"Blue"]),
                (b"TE", [b"things"]),
                (b"WWW-Authenticate", [b"hooray"]),
                (b"X-XSS-Protection", [b"done"]),
                (b"Yellow-And-Green", [b"purple-too"]),
            ],
        )

    def test_eq(self):
        Headers = self.Headers
        self.assertTrue(
            Headers([(b"foo", [b"bar"])]) == Headers([(b"foo", [b"bar"])])
        )
        self.assertFalse(
            Headers([(b"foo", [b"bar"])]) == Headers([(b"foo", [])])
        )

    def test_ne(self):
        Headers = self.Headers
        self.assertTrue(
            Headers([(b"foo", [b"bar"])]) != Headers([(b"foo", [])])
        )
        self.assertFalse(
            Headers([(b"foo", [b"bar"])]) != Headers([(b"foo", [b"bar"])])
        )

    def test_eq_NotImplemented(self):
        self.assertFalse(self.Headers() == {})

    def test_ne_NotImplemented(self):
        self.assertTrue(self.Headers() != {})

    def test_repr(self):
        headers = self.Headers(
            [(b"foo", [b"bar"]), (b"thing", [b"baz", b"quux"])]
        )

        if PY3:
            fields = " contents={b'Foo'=[b'bar'] b'Thing'=[b'baz', b'quux']}>"
        else:
            fields = " contents={Foo=['bar'] Thing=['baz', 'quux']}>"

        self.assertEqual(repr(headers), "<" + self.Headers.__name__ + fields)

    def test_empty_repr(self):
        headers = self.Headers()
        self.assertEqual(
            repr(headers), "<" + self.Headers.__name__ + " contents={}>",
        )


class TestMutableHeaders(HeaderRetrievalTestsMixin, TestCase):

    Headers = http.MutableHeaders

    def test_mutated_contains(self):
        headers = self.Headers()
        self.assertNotIn(b"thing", headers)

        headers[b"Thing"] = [b"hello", b"world"]

        self.assertIn(b"thing", headers)
        self.assertEqual(headers.get(b"thing"), [b"hello", b"world"])

    def test_add_value(self):
        headers = self.Headers([(b"foo", [b"bar"])])
        headers.add_value(b"foo", b"baz")
        self.assertEqual(headers.get(b"foo"), [b"bar", b"baz"])

    def test_add_new_value(self):
        headers = self.Headers()
        headers.add_value(b"thing", b"value")
        self.assertEqual(headers.get(b"thing"), [b"value"])

    def test_pop(self):
        headers = self.Headers([(b"thing", [b"value"])])
        self.assertEqual(headers.pop(b"thing"), [b"value"])
        self.assertNotIn(b"thing", headers)

    def test_pop_with_default_returning_default(self):
        headers = self.Headers()
        default = object()
        self.assertIs(headers.pop(b"nonexisting", default), default)

    def test_pop_with_default_returning_value(self):
        headers = self.Headers([(b"existing", [b"value"])])
        default = object()
        self.assertEqual(headers.pop(b"existing", default), [b"value"])
        self.assertNotIn(b"existing", headers)

    def test_pop_defaults_to_a_lookup_error(self):
        headers = self.Headers()
        with self.assertRaises(http.NoSuchHeader):
            self.assertIsNone(headers.pop(b"nonexisting"))

    def test_mutated_eq_ne(self):
        first = self.Headers([(b"foo", [b"bar"])])
        second = self.Headers()

        second.add_value(b"foo", b"bar")
        self.assertEqual(first, second)

        second.add_value(b"foo", b"baz")
        self.assertNotEqual(first, second)

    def test_not_hashable(self):
        with self.assertRaises(TypeError):
            hash(self.Headers())


class TestImmutableHeaders(HeaderRetrievalTestsMixin, TestCase):

    Headers = http.Headers

    def test_hash(self):
        some_headers = set(
            [self.Headers([(b"a", [b"b"])]), self.Headers([(b"a", [b"b"])])],
        )
        self.assertEqual(some_headers, set([self.Headers([(b"a", [b"b"])])]))

    def test_empty_hash(self):
        some_headers = set([self.Headers(), self.Headers()])
        self.assertEqual(some_headers, set([self.Headers()]))


class TestAccept(TestCase):
    def test_basic(self):
        accept = http.Accept.from_header(header=b"application/json")
        media_range = http.MediaRange(
            type=b"application", subtype=b"json", quality=1.0,
        )
        self.assertEqual(accept, http.Accept(media_types=(media_range,)))

    def test_multiple(self):
        accept = http.Accept.from_header(header=b"audio/*; q=0.2, audio/basic")
        media_types = (
            http.MediaRange(type=b"audio", quality=0.2),
            http.MediaRange(type=b"audio", subtype=b"basic", quality=1.0),
        )
        self.assertEqual(accept, http.Accept(media_types=media_types))

    def test_strips_spaces(self):
        accept = http.Accept.from_header(header=b" text/* ; q=0.2 ,  text/foo")
        media_types = (
            http.MediaRange(type=b"text", quality=0.2),
            http.MediaRange(type=b"text", subtype=b"foo", quality=1.0),
        )
        self.assertEqual(accept, http.Accept(media_types=media_types))

    def test_more_elaborate(self):
        accept = http.Accept.from_header(
            header=b"text/plain; q=0.5, text/html, text/x-dvi; q=0.8, text/x-c"
        )
        media_types = (
            http.MediaRange(type=b"text", subtype=b"plain", quality=0.5),
            http.MediaRange(type=b"text", subtype=b"x-dvi", quality=0.8),
            http.MediaRange(type=b"text", subtype=b"html"),
            http.MediaRange(type=b"text", subtype=b"x-c"),
        )
        self.assertEqual(accept, http.Accept(media_types=media_types))

    def test_override_with_more_specific_type(self):
        accept = http.Accept.from_header(
            header=b"text/*, text/plain, text/plain;format=flowed, */*",
        )
        media_types = (
            http.MediaRange(),
            http.MediaRange(type=b"text"),
            http.MediaRange(type=b"text", subtype=b"plain"),
            http.MediaRange(
                type=b"text",
                subtype=b"plain",
                parameters=pmap({b"format": b"flowed"}),
            ),
        )
        self.assertEqual(accept, http.Accept(media_types=media_types))

    def test_quality_factors(self):
        accept = http.Accept.from_header(
            header=(
                b"text/*;q=0.3, text/html;q=0.7, text/html;level=1, "
                b"text/html;level=2;q=0.4, */*;q=0.5"
            ),
        )
        media_types = (
            http.MediaRange(type=b"text", quality=0.3),
            http.MediaRange(
                type=b"text",
                subtype=b"html",
                quality=0.4,
                parameters=pmap({b"level": b"2"}),
            ),
            http.MediaRange(quality=0.5),
            http.MediaRange(type=b"text", subtype=b"html", quality=0.7),
            http.MediaRange(
                type=b"text",
                subtype=b"html",
                parameters=pmap({b"level": b"1"}),
            ),
        )
        self.assertEqual(accept, http.Accept(media_types=media_types))

    def test_no_header(self):
        accept = http.Accept.from_header(header=None)
        self.assertEqual(accept, http.Accept.ALL)


class TestMediaRange(TestCase):
    def test_eq(self):
        self.assertTrue(
            http.MediaRange(type="text", subtype="plain", quality=0.5) ==
            http.MediaRange(type="text", subtype="plain", quality=0.5),
        )

    def test_eq_type(self):
        self.assertFalse(
            http.MediaRange(type="text", subtype="plain", quality=0.5) ==
            http.MediaRange(type="foo", subtype="plain", quality=0.5),
        )

    def test_eq_subtype(self):
        self.assertFalse(
            http.MediaRange(type="text", subtype="plain", quality=0.5) ==
            http.MediaRange(type="text", subtype="other", quality=0.5),
        )

    def test_eq_quality(self):
        self.assertFalse(
            http.MediaRange(type="text", subtype="plain", quality=0.5) ==
            http.MediaRange(type="text", subtype="plain", quality=0.8),
        )

    def test_eq_parameters(self):
        self.assertFalse(
            http.MediaRange(type="text", parameters={"a": "b"}) ==
            http.MediaRange(type="text", parameters={"a": "c"}),
        )

    def test_ne(self):
        self.assertFalse(
            http.MediaRange(type="text", subtype="plain", quality=0.5) !=
            http.MediaRange(type="text", subtype="plain", quality=0.5),
        )

    def test_ne_type(self):
        self.assertTrue(
            http.MediaRange(type="text", subtype="plain", quality=0.5) !=
            http.MediaRange(type="foo", subtype="plain", quality=0.5),
        )

    def test_ne_subtype(self):
        self.assertTrue(
            http.MediaRange(type="text", subtype="plain", quality=0.5) !=
            http.MediaRange(type="text", subtype="other", quality=0.5),
        )

    def test_ne_quality(self):
        self.assertTrue(
            http.MediaRange(type="text", subtype="plain", quality=0.5) !=
            http.MediaRange(type="text", subtype="plain", quality=0.8),
        )

    def test_ne_parameters(self):
        self.assertTrue(
            http.MediaRange(type="text", parameters={"a": "b"}) !=
            http.MediaRange(type="text", parameters={"a": "c"}),
        )

    def test_lt_same_type_and_subtype(self):
        self.assertLess(
            http.MediaRange(type="text", subtype="plain", quality=0.5),
            http.MediaRange(type="text", subtype="plain", quality=0.8),
        )

    def test_lt_ranged_subtype(self):
        self.assertLess(
            http.MediaRange(type="text"),
            http.MediaRange(type="text", subtype="plain"),
        )

    def test_lt_ranged_type_and_subtype(self):
        self.assertLess(http.MediaRange(), http.MediaRange(type="text"))

    def test_lt_ranged_lower_quality(self):
        self.assertLess(
            http.MediaRange(type="text", quality=0.3),
            http.MediaRange(quality=0.5),
        )

    def test_lt_different_type_and_subtype(self):
        self.assertLess(
            http.MediaRange(type="application", subtype="json", quality=0.5),
            http.MediaRange(type="text", subtype="plain", quality=0.8),
        )

    def test_lt_parameters_same_type_and_subtype(self):
        self.assertLess(
            http.MediaRange(type="bar", subtype="foo"),
            http.MediaRange(
                type="bar",
                subtype="foo",
                parameters=pmap({"a": "b"}),
            ),
        )

    def test_not_lt_same_type_and_subtype(self):
        self.assertFalse(
            http.MediaRange(type="text", subtype="plain", quality=0.8) <
            http.MediaRange(type="text", subtype="plain", quality=0.5),
        )

    def test_not_lt_different_type_and_subtype(self):
        self.assertFalse(
            http.MediaRange(type="application", subtype="json") <
            http.MediaRange(type="text", subtype="plain"),
        )

    def test_not_lt_parameters(self):
        self.assertFalse(
            http.MediaRange(
                type="bar",
                subtype="foo",
                parameters=pmap({"a": "b"}),
            ) < http.MediaRange(type="bar", subtype="foo"),
        )

    def test_not_lt_parameters_different_type(self):
        self.assertFalse(
            http.MediaRange(type="foo", subtype="foo") < http.MediaRange(
                type="bar",
                subtype="foo",
                parameters=pmap({"a": "b"}),
            ),
        )

    def test_not_lt_parameters_different_subtype(self):
        self.assertFalse(
            http.MediaRange(type="bar", subtype="foo") < http.MediaRange(
                type="bar",
                subtype="baz",
                parameters=pmap({"a": "b"}),
            ),
        )

    def test_not_lt_ranged_subtype(self):
        self.assertFalse(
            http.MediaRange(type="text", subtype="plain") <
            http.MediaRange(type="text"),
        )

    def test_not_lt_ranged_lower_quality(self):
        self.assertFalse(
            http.MediaRange(quality=0.5) <
            http.MediaRange(type="text", quality=0.3),
        )

    def test_lt_equal_range(self):
        self.assertFalse(http.MediaRange() < http.MediaRange())

    def test_hash(self):
        self.assertEqual(
            {
                http.MediaRange(type="text", subtype="plain", quality=0.8),
                http.MediaRange(type="text", subtype="plain", quality=0.8),
            },
            {http.MediaRange(type="text", subtype="plain", quality=0.8)},
        )

    def test_hash_parameters(self):
        self.assertEqual(
            {
                http.MediaRange(type="a", parameters=pmap({"a": "b"})),
                http.MediaRange(type="a", parameters=pmap({"a": "b"})),
            },
            {http.MediaRange(type="a", parameters=pmap({"a": "b"}))},
        )

    def test_hash_different_parameters(self):
        self.assertNotEqual(
            hash(http.MediaRange(type="a", parameters=pmap({"a": "b"}))),
            hash(http.MediaRange(type="a", parameters=pmap({"a": "c"}))),
        ),



_NOTHING_YET = object()


class Deferred(object):
    """
    A coordinator of a future result.

    """

    def __init__(self):
        chain = _CallbackChain()
        self.chain = chain
        self.succeed = chain._succeed

    def error(self, result):
        """
        Report having failed with the given error.

        """

        self.chain._error(result=result)


class _CallbackChain(object):
    """
    A callback chain manages a collection of success- and error- handlers.

    """

    _resulted_in = _NOTHING_YET

    def __init__(self):
        self._callbacks = []

    def on_success(self, fn, *args, **kwargs):
        """
        Call the given callback if or when the connected deferred succeeds.

        """

        self._callbacks.append((fn, args, kwargs))

        result = self._resulted_in
        if result is not _NOTHING_YET:
            self._succeed(result=result)

    def on_error(self, fn, *args, **kwargs):
        """
        Call the given callback if or when the connected deferred errors.

        """

    def _succeed(self, result):
        """
        Fire the success chain.

        """

        for fn, args, kwargs in self._callbacks:
            fn(result, *args, **kwargs)
        self._resulted_in = result



from unittest import TestCase, skipIf
import json

from future.utils import PY3
from hyperlink import URL

from minion import routing
from minion.request import Request, Response, redirect
from minion.traversal import LeafResource


class ReverseRenderer(object):
    def render(self, request, response):
        return Response(b"".join(reversed(response.content)))


def view(request, **kwargs):
    return Response(json.dumps(kwargs))


class TestRouter(TestCase):
    def setUp(self):
        self.router = routing.Router(mapper=routing.SimpleMapper())

    def test_route(self):
        self.router.add(b"/", view)
        request = Request(url=URL(path=[u""]))
        response = self.router.route(request, path=b"/")
        self.assertEqual(response, view(request))

    def test_unknown_route(self):
        request = Request(url=URL(path=[u"404"]))
        response = self.router.route(request, path=b"/404")
        self.assertEqual(response, Response(code=404))

    def test_specified_renderer(self):
        self.router.add(b"/", view, renderer=ReverseRenderer())
        request = Request(url=URL(path=[u""]))
        response = self.router.route(request, path=b"/")
        self.assertEqual(response, Response(b"}{"))

    def test_specified_renderer_multiple_routes(self):
        self.router.add(b"/", view, methods=[b"GET"], renderer=None)
        self.router.add(
            b"/", view, methods=[b"POST"], renderer=ReverseRenderer(),
        )

        get = Request(url=URL(path=[u""]), method=b"GET")
        response = self.router.route(get, path=b"/")
        self.assertEqual(response, Response(b"{}"))

        post = Request(url=URL(path=[u""]), method=b"POST")
        response = self.router.route(post, path=b"/")
        self.assertEqual(response, Response(b"}{"))

    def test_renderer_with_render_error_handler(self):
        class RendererWithErrorHandler(object):
            def render(self, request, response):
                raise ZeroDivisionError()

            def render_error(self, request, response, exc):
                return Response(response.content + exc.__class__.__name__)

        self.router.add(b"/", view, renderer=RendererWithErrorHandler())
        response = self.router.route(Request(url=URL(path=[u""])), path=b"/")
        self.assertEqual(response, Response(b"{}ZeroDivisionError"))

    def test_renderer_without_error_handler(self):
        class RendererWithoutErrorHandler(object):
            def render(self, request, response):
                raise ZeroDivisionError()

        self.router.add(b"/", view, renderer=RendererWithoutErrorHandler())
        request = Request(url=URL(path=[u""]))

        with self.assertRaises(ZeroDivisionError):
            self.router.route(request, path=b"/")

    def test_renderer_with_view_error_handler(self):
        class RendererWithErrorHandler(object):
            def render(self, request, response):
                raise ZeroDivisionError("I won't ever get raised")

            def view_error(self, request, error):
                return Response(error.__class__.__name__)

        def boom(request):
            raise IndexError()

        self.router.add(b"/", boom, renderer=RendererWithErrorHandler())
        response = self.router.route(Request(url=URL(path=[u""])), path=b"/")
        self.assertEqual(response, Response(b"IndexError"))


class TestRouterDefaultRenderer(TestCase):
    def setUp(self):
        self.router = routing.Router(
            mapper=routing.SimpleMapper(), default_renderer=ReverseRenderer(),
        )

    def test_default_renderer(self):
        self.router.add(b"/", view)
        request = Request(url=URL(path=[u""]))
        response = self.router.route(request, path=b"/")
        self.assertEqual(response, Response(b"}{"))

    def test_disable_default_renderer(self):
        self.router.add(b"/", view, renderer=None)
        request = Request(url=URL(path=[u""]))
        response = self.router.route(request, path=b"/")
        self.assertEqual(response, Response(b"{}"))


class MapperTestMixin(object):
    """
    Test basic functionality common to all mappers.

    Expects a mapper to be present and instantiated at the `mapper` attribute
    of the test case.

    """

    def test_it_maps_routes(self):
        self.mapper.add(b"/route", view)
        render = self.mapper.map(
            Request(url=URL(path=[u"route"], rooted=True)),
            path=b"/route",
        )
        self.assertIs(render, view)

    def test_it_maps_routes_for_specified_methods(self):
        self.mapper.add(b"/route", view, methods=[b"POST"])
        request = Request(
            url=URL(path=[u"route"], rooted=True),
            method=b"POST",
        )
        self.assertEqual(self.mapper.map(request, path=b"/route"), view)

    def test_it_does_not_map_routes_for_unspecified_methods(self):
        self.mapper.add(b"/route", view, methods=[b"POST"])
        request = Request(url=URL(path=[u"route"]), method=b"GET")
        self.assertIsNone(self.mapper.map(request, path=b"/route"))

    def test_it_maps_multiple_routes_for_specified_methods(self):
        self.mapper.add(b"/", lambda r: Response(b"GET"), methods=[b"GET"])
        self.mapper.add(b"/", lambda r: Response(b"POST"), methods=[b"POST"])

        get = Request(url=URL(path=[u""], rooted=True), method=b"GET")
        post = Request(url=URL(path=[u""], rooted=True), method=b"POST")
        map = self.mapper.map
        self.assertEqual(
            (map(get, path=b"/")(get), map(post, path=b"/")(post)),
            (Response(b"GET"), Response(b"POST")),
        )

    def test_it_can_build_named_routes(self):
        self.mapper.add(b"/", view, route_name=u"home")
        self.assertEqual(self.mapper.lookup(u"home"), b"/")

    def test_unknown_route_names_become_literal_paths(self):
        self.assertEqual(self.mapper.lookup(b"/work"), b"/work")

    def test_it_does_not_map_unknown_paths(self):
        request = Request(url=URL(path=[u"route"]))
        render = self.mapper.map(request, path=b"/route")
        self.assertIsNone(render)

    def test_extra_build_arguments_become_query_strings(self):
        self.mapper.add(b"/", view, route_name=u"home")
        url = self.mapper.lookup(u"home", thing=b"yes")
        self.assertEqual(url, b"/?thing=yes")


class TestTraversalMapper(TestCase):
    def test_it_traverses_children(self):
        class Resource(object):
            def get_child(self, name, request):
                return LeafResource(
                    render=lambda request: Response(b"Hello " + name),
                )

        mapper = routing.TraversalMapper(root=Resource())
        request = Request(url=URL(path=[u"world"]))
        render = mapper.map(request, path=b"/world")
        self.assertEqual(render(request), Response(content=b"Hello world"))


class TestTraversalMapperStatic(MapperTestMixin, TestCase):
    def setUp(self):
        self.mapper = routing.TraversalMapper(root=LeafResource(render=None))


@skipIf(not hasattr(routing, "RoutesMapper"), "Routes not found")
@skipIf(PY3, "WSGI on Py3 is insanity")
class TestRoutesMapper(MapperTestMixin, TestCase):
    def setUp(self):
        self.mapper = routing.RoutesMapper()

    def test_it_maps_routes_with_arguments(self):
        self.mapper.add(b"/route/{year}", view, stuff=b"12")
        request = Request(url=URL(path=[u"route", u"2013"], rooted=True))
        render = self.mapper.map(request, path=b"/route/2013")
        self.assertEqual(
            json.loads(render(request).content),
            {b"year": b"2013", b"stuff": b"12"},
        )

    def test_it_builds_routes_with_arguments(self):
        self.mapper.add(b"/{year}", view, route_name=u"year")
        url = self.mapper.lookup(b"year", year=2012)
        self.assertEqual(url, b"/2012")


@skipIf(not hasattr(routing, "WerkzeugMapper"), "Werkzeug not found")
@skipIf(PY3, "WSGI on Py3 is insanity")
class TestWerkzeugMapper(MapperTestMixin, TestCase):
    def setUp(self):
        self.mapper = routing.WerkzeugMapper()

    def test_it_maps_routes_with_arguments(self):
        self.mapper.add(b"/route/<int:year>", view)
        request = Request(url=URL(path=[u"route", u"2013"], rooted=True))
        render = self.mapper.map(request, path=b"/route/2013")
        self.assertEqual(json.loads(render(request).content), {b"year": 2013})

    def test_it_builds_routes_with_arguments(self):
        self.mapper.add(b"/<int:year>", view, route_name=u"year")
        url = self.mapper.lookup(u"year", year=2012)
        self.assertEqual(url, b"/2012")

    def test_it_handles_routing_redirects(self):
        self.mapper.add(b"/<int:year>/", view)
        request = Request(url=URL(path=[u"2013"], rooted=True))
        render = self.mapper.map(request, path=b"/2013")
        self.assertEqual(
            render(request), redirect(b"http:///2013/", code=301),
        )


class TestSimpleMapper(MapperTestMixin, TestCase):
    def setUp(self):
        self.mapper = routing.SimpleMapper()



"""
Routing building blocks.

There are two fundamental objects responsible for routing:

    * Mappers map a URL (an instance of :class:`bytes`) to a view (a
        callable) with some optional additional arguments
    * Routers coordinate the entire routing process, taking a request
        and returning a response, typically by composing with and invoking
        one or more Mappers.

This module defines (different) implementations which can be combined to
perform these operations in various ways.

"""

from collections import defaultdict
from functools import partial

from future.utils import listitems as items
from future.moves.urllib.parse import urlencode
import attr

from minion.renderers import bind
from minion.request import Response, redirect
from minion.traversal import traverse


@attr.s
class Router(object):

    mapper = attr.ib()
    default_renderer = attr.ib(default=None)

    def add(
        self,
        route,
        fn,
        renderer=False,
        route_name=None,
        methods=(b"GET", b"HEAD"),
        **kw
    ):
        if renderer is False:
            renderer = self.default_renderer
        if renderer is not None:
            fn = bind(renderer=renderer, to=fn)
        self.mapper.add(
            route, fn, route_name=route_name, methods=methods, **kw
        )

    def route(self, request, path):
        render = self.mapper.map(request=request, path=path)
        if render is not None:
            response = render(request=request)
        else:
            response = Response(code=404)
        return response


try:
    import routes
except ImportError:
    pass
else:
    class RoutesMapper(object):
        """
        A mapper that maps via `routes <http://routes.readthedocs.org/>`_.

        """

        def __init__(self, mapper=None):
            if mapper is None:
                mapper = routes.Mapper()
            self._generator = routes.URLGenerator(mapper, {})  # XXX: environ
            self._mapper = mapper

        def add(self, route, fn, route_name=None, methods=None, **kwargs):
            if methods is not None:
                kwargs.setdefault("conditions", {})["method"] = methods
            self._mapper.connect(route_name, route, minion_target=fn, **kwargs)

        def map(self, request, path):
            match = self._mapper.match(
                request.url.to_text(),
                # Yes seriously. This seems to be the only way to do this.
                environ={"REQUEST_METHOD": request.method},
            )
            if match is None:
                return None

            render = match.pop("minion_target")
            if match:
                render = partial(render, **match)
            return render

        def lookup(self, route_name, **kwargs):
            return self._generator(route_name, **kwargs)


try:
    import werkzeug.routing
except ImportError:
    pass
else:
    class WerkzeugMapper(object):
        """
        A mapper that uses Werkzeug's routing.

        """

        def __init__(self, map=None):
            if map is None:
                map = werkzeug.routing.Map()
            self._endpoints = {}
            self._map = map
            self._adapter = self._map.bind(b"")  # XXX: server_name

        def add(self, route, fn, route_name=None, methods=None, **kwargs):
            if methods is not None:
                kwargs["methods"] = methods
            endpoint = route_name or fn
            self._endpoints[endpoint] = fn
            rule = werkzeug.routing.Rule(route, endpoint=endpoint, **kwargs)
            self._map.add(rule)

        def map(self, request, path):
            try:
                render, kwargs = self._adapter.match(
                    path_info=request.url.to_text(),
                    method=request.method,
                )
            except werkzeug.routing.RequestRedirect as redirect_exception:
                return lambda request: redirect(
                    to=redirect_exception.new_url,
                    code=redirect_exception.code,
                )
            except werkzeug.routing.HTTPException:
                return None
            else:
                if kwargs:
                    render = partial(render, **kwargs)
                return render

        def lookup(self, route_name, **kwargs):
            try:
                return self._adapter.build(route_name, kwargs)
            except werkzeug.routing.BuildError:
                return route_name


class TraversalMapper(object):
    """
    Object-traversal based mapper for traversal of resource objects.

    """

    def __init__(self, root, static_mapper=None):
        if static_mapper is None:
            static_mapper = SimpleMapper()
        self.root = root

        self.static_mapper = static_mapper
        self.add = static_mapper.add
        self.lookup = static_mapper.lookup

    def map(self, request, path):
        static_render = self.static_mapper.map(request=request, path=path)
        if static_render is not None:
            return static_render
        resource = traverse(path=path, request=request, resource=self.root)
        return resource.render


class SimpleMapper(object):
    """
    Simple dictionary based lookup-routing without parameters.

    """

    def __init__(self):
        self._routes = defaultdict(dict)
        self._names = {}

    def add(self, route, fn, route_name=None, methods=(b"GET", b"HEAD")):
        for method in methods:
            self._routes[method][route] = fn
        if route_name is not None:
            self._names[route_name] = route

    def map(self, request, path):
        return self._routes[request.method].get(path)

    def lookup(self, route_name, **kwargs):
        url = self._names.get(route_name, route_name)
        if kwargs:
            url += b"?" + urlencode(items(kwargs))
        return url



from unittest import TestCase
import json

from hyperlink import URL

from minion.core import Application
from minion.http import Headers
from minion.renderers import JSON
from minion.request import Request, Response
from minion.routing import Router, SimpleMapper


class TestMinion(TestCase):
    def test_it_routes_simple_views(self):
        minion = Application(router=Router(mapper=SimpleMapper()))

        @minion.route(b"/show")
        def show(request):
            return Response(b"Hello World!")

        url = URL(path=[u"show"])
        response = minion.serve(Request(url=url), path=u"/show")
        self.assertEqual(response, Response(b"Hello World!"))


class RequestIntegrationTestMixin(object):
    def setUp(self):
        self.minion = Application()

        @self.minion.route(b"/respond", renderer=JSON())
        def respond(request):
            return {"url": request.url.to_text().encode("utf-8")}

    def get_request(self, *args, **kwargs):
        kwargs.setdefault("headers", Headers())
        response = json.loads(self.get(*args, **kwargs))
        response["url"] = URL.from_text(response["url"].decode("utf-8"))
        return Request(**response)

    def test_it_parses_the_url(self):
        request = self.get_request(
            b"/respond?foo=bar#baz",
            headers=Headers([(b"Host", [b"example.com"])]),
        )
        self.assertEqual(
            request.url, URL(
                scheme=u"http",
                host=u"example.com",
                path=[u"respond"],
                query=[(u"foo", u"bar")],
                fragment=u"",  # Fragments should be ignored by servers.
            ),
        )



import os

from setuptools import find_packages, setup


with open(os.path.join(os.path.dirname(__file__), "README.rst")) as readme:
    long_description = readme.read()

classifiers = [
    "Development Status :: 3 - Alpha",
    "License :: OSI Approved :: MIT License",
    "Operating System :: OS Independent",
    "Programming Language :: Python",
    "Programming Language :: Python :: 2.7",
    "Programming Language :: Python :: 3.4",
    "Programming Language :: Python :: 2",
    "Programming Language :: Python :: 3",
    "Programming Language :: Python :: Implementation :: CPython",
    "Programming Language :: Python :: Implementation :: PyPy",
]

setup(
    name="minion",
    packages=find_packages(),
    setup_requires=["vcversioner"],
    install_requires=[
        "attrs",
        "cached-property",
        "future",
        "itsdangerous",
        "pyrsistent",
        "Werkzeug",
    ],
    author="Julian Berman",
    author_email="Julian@GrayVines.com",
    classifiers=classifiers,
    license="MIT",
    long_description=long_description,
    url="https://github.com/Julian/Minion",
    description="A microframework based on evil intentions and "
                "whatever else you've got",
    vcversioner={"version_module_paths" : ["minion/_version.py"]},
)



from __future__ import absolute_import
from io import BytesIO

from cached_property import cached_property as calculated_once
from future.utils import iteritems
from werkzeug.http import HTTP_STATUS_CODES as HTTP_STATUS_PHRASES
import attr

from minion.deferred import Deferred
from minion.http import Accept, Headers, MutableHeaders


HTTP_STATUS_CODES = dict(
    (code, "{0} {1}".format(code, phrase))
    for code, phrase in iteritems(HTTP_STATUS_PHRASES)
)


class Responder(object):
    """
    A responder represents a pending HTTP response for a corresponding request.

    """

    def __init__(self, request):
        self._after_deferreds = []
        self.request = request

    def after(self):
        """
        Return a deferred that will fire after the request is finished.

        Returns:

            Deferred: a new deferred that will fire appropriately

        """

        d = Deferred()
        self._after_deferreds.append(d)
        return d.chain

    def finish(self):
        for d in self._after_deferreds:
            d.succeed(self)


class Manager(object):
    """
    The request manager coordinates state during each active request.

    """

    def __init__(self):
        self._requests = {}

    def after_response(self, request, fn, *args, **kwargs):
        """
        Call the given callable after the given request has its response.

        Arguments:

            request:

                the request to piggyback

            fn (callable):

                a callable that takes at least two arguments, the request and
                the response (in that order), along with any additional
                positional and keyword arguments passed to this function which
                will be passed along. If the callable returns something other
                than ``None``, it will be used as the new response.
        """

        self._requests[id(request)]["callbacks"].append((fn, args, kwargs))

    def request_started(self, request):
        self._requests[id(request)] = {"callbacks": [], "assets": {}}

    def request_served(self, request, response):
        request_data = self._requests.pop(id(request))
        for callback, args, kwargs in request_data["callbacks"]:
            callback_response = callback(response, *args, **kwargs)
            if callback_response is not None:
                response = callback_response
        return response


@attr.s
class Request(object):

    url = attr.ib()
    content = attr.ib(default=attr.Factory(BytesIO))
    headers = attr.ib(default=attr.Factory(Headers))
    method = attr.ib(default=b"GET")
    messages = attr.ib(default=attr.Factory(list), cmp=False)

    @calculated_once
    def accept(self):
        header = self.headers.get("Accept")
        if header is not None:
            header = ",".join(header)
        return Accept.from_header(header=header)

    def flash(self, message):
        self.messages.append(_Message(content=message))


@attr.s
class Response(object):

    content = attr.ib(default="")
    code = attr.ib(default=200)
    headers = attr.ib(default=attr.Factory(MutableHeaders))

    @property
    def status(self):
        return HTTP_STATUS_CODES[self.code]


@attr.s
class _Message(object):
    """
    A flashed message.

    """

    content = attr.ib()


def redirect(to, code=302):
    """
    Return a redirect response to the given URL.

    """

    return Response(headers=MutableHeaders([("Location", [to])]), code=code)



"""
Helpers for applications that use object traversal routing.

"""

from future.utils import PY3, iteritems
import attr

from minion.request import Response


@attr.s
class LeafResource(object):
    """
    A leaf resource that simply renders via a provided view.

    """

    is_leaf = True
    render = attr.ib()


@attr.s
class TreeResource(object):
    """
    A tree resource that supports adding children via :meth:`set_child`\ .

    """

    render = attr.ib()
    _children = attr.ib(default=attr.Factory(dict))

    _no_such_child = LeafResource(render=lambda request: Response(code=404))

    def get_child(self, name, request):
        child = self._children.get(name)
        if child is None:
            return self._no_such_child
        return child

    def set_child(self, name, resource):
        self._children[name] = resource


def method_delegate(**methods):
    """
    Construct a renderer that delegates based on the request's HTTP method.

    """

    methods = {k.upper(): v for k, v in iteritems(methods)}

    if PY3:
        methods = {k.encode("utf-8"): v for k, v in iteritems(methods)}

    def render(request):
        renderer = methods.get(request.method)
        if renderer is None:
            return Response(code=405)
        return renderer(request)
    return render


def traverse(path, request, resource):
    """
    Traverse a root resource, retrieving the appropriate child for the request.

    """

    path = path.lstrip(b"/")
    for component in path and path.split(b"/"):
        if getattr(resource, "is_leaf", False):
            break
        resource = resource.get_child(name=component, request=request)
    return resource



# -*- coding: utf-8 -*-
from collections import OrderedDict
from unittest import TestCase
import json

from hyperlink import URL

from minion import renderers
from minion.http import Headers
from minion.request import Request, Response


class TestUnicodeRenderer(TestCase):

    request = Request(url=URL(path=[u""]))

    def test_it_renders_via_the_given_encoding(self):
        renderer = renderers.Unicode(encoding="utf-8")
        render = renderers.bind(renderer, to=lambda _: u"שלום")
        self.assertEqual(
            render(self.request), Response(u"שלום".encode("utf-8")),
        )

    def test_encoding_ignoring_errors(self):
        renderer = renderers.Unicode(encoding="ascii", errors="ignore")
        render = renderers.bind(renderer, to=lambda _: u"שלום")
        self.assertEqual(render(self.request), Response(b""))

    def test_encoding_errors_by_default(self):
        renderer = renderers.Unicode(encoding="ascii")
        render = renderers.bind(renderer, to=lambda _: u"שלום")
        with self.assertRaises(UnicodeEncodeError):
            render(self.request)

    def test_UTF8(self):
        render = renderers.bind(renderers.UTF8, to=lambda _: u"שלום")
        self.assertEqual(
            render(self.request), Response(u"שלום".encode("utf-8")),
        )


class TestJSON(TestCase):
    def assertPretty(self, content, response):
        self.assertNotEqual(
            response, Response(
                headers=Headers([("Content-Type", ["application/json"])]),
                content=json.dumps(content),
            ),
        )
        self.assertEqual(
            response, Response(
                headers=Headers([("Content-Type", ["application/json"])]),
                content=json.dumps(content, indent=2, sort_keys=True)
            ),
        )

    def assertNotPretty(self, content, response):
        self.assertEqual(
            response, Response(
                headers=Headers([("Content-Type", ["application/json"])]),
                content=json.dumps(content, separators=",:"),
            ),
        )
        self.assertNotEqual(
            response, Response(
                headers=Headers([("Content-Type", ["application/json"])]),
                content=json.dumps(content, indent=2, sort_keys=True)
            ),
        )

    def test_it_dumps_pretty_json_for_humans(self):
        content = ["a", "b", "c"]
        render = renderers.bind(renderers.JSON(), to=lambda _: content)
        request = Request(
            url=URL(path=[u""]),
            headers=Headers([("Accept", ["*/*"])]),
        )
        self.assertPretty(content, render(request))

    def test_it_dumps_pretty_json_for_humans_no_accept_header(self):
        content = ["a", "b", "c"]
        render = renderers.bind(renderers.JSON(), to=lambda _: content)
        self.assertPretty(content, render(Request(url=URL(path=[u""]))))

    def test_it_dumps_practical_json_for_machines(self):
        content = ["a", "b", "c"]
        render = renderers.bind(renderers.JSON(), to=lambda _: content)
        request = Request(
            url=URL(path=[u""]),
            headers=Headers([("Accept", ["application/json"])]),
        )
        self.assertNotPretty(content, render(request))

    def test_dict_sort_keys(self):
        content = OrderedDict([("foo", "bar"), ("baz", "quux")])
        render = renderers.bind(renderers.JSON(), to=lambda _: content)
        request = Request(
            url=URL(path=[u""]),
            headers=Headers([("Accept", ["*/*"])]),
        )
        self.assertPretty(content, render(request))

    def test_separators(self):
        render = renderers.bind(renderers.JSON(), to=lambda _: dict(a=1, b=2))
        request = Request(
            url=URL(path=[u""]),
            headers=Headers([("Accept", ["application/json"])]),
        )
        response = render(request)
        self.assertEqual(
            response, Response(
                content=response.content,
                headers=Headers([("Content-Type", ["application/json"])]),
            ),
        )
        self.assertNotIn(response.content, " ")
        self.assertEqual(json.loads(response.content), {"a": 1, "b": 2})

    def test_customized_dumps(self):
        """
        A customized dumps applies to both pretty and non-pretty responses.

        """

        renderer = renderers.JSON(default=lambda obj: 23)
        render = renderers.bind(renderer, to=lambda _: {"foo": object()})
        request = Request(
            url=URL(path=[u""]),
            headers=Headers([("Accept", ["application/json"])]),
        )
        self.assertEqual(
            render(request), Response(
                content=b'{"foo":23}',
                headers=Headers([("Content-Type", ["application/json"])]),
            ),
        )
        self.assertEqual(
            render(Request(url=URL(path=[u""]))), Response(
                content=b'{\n  "foo": 23\n}',
                headers=Headers([("Content-Type", ["application/json"])]),
            ),
        )


class TestSimpleJSON(TestCase):

    request = Request(url=URL(path=[u""]))

    def test_it_dumps_json(self):
        renderer = renderers.SimpleJSON()
        render = renderers.bind(renderer, to=lambda _: {"foo": "bar"})
        self.assertEqual(render(self.request), Response(b'{"foo": "bar"}'))

    def test_customized_dumps(self):
        renderer = renderers.SimpleJSON(separators=",:")
        render = renderers.bind(renderer, to=lambda _: {"foo": "bar"})
        self.assertEqual(render(self.request), Response(b'{"foo":"bar"}'))



from unittest import TestCase

from minion.http import Headers
from minion.request import Response
from minion.tests import utils


class TestResponseHelpersMixin(utils.ResponseHelpersMixin, TestCase):
    def test_assertRedirected_succeeds_for_redirects(self):
        response = Response(
            headers=Headers([("Location", ["http://example.com"])]), code=302,
        )
        self.assertRedirected(response, to="http://example.com", code=302)

    def test_assertRedirected_fails_for_mismatched_urls(self):
        response = Response(
            headers=Headers([("Location", ["http://example.com"])]), code=302,
        )
        with self.assertRaises(self.failureException):
            self.assertRedirected(response, to="http://other.com", code=302)

    def test_assertRedirected_fails_for_mismatched_code(self):
        response = Response(
            headers=Headers([("Location", ["http://example.com"])]), code=301,
        )
        with self.assertRaises(self.failureException):
            self.assertRedirected(response, to="http://example.com", code=302)

    def test_assertRedirected_fails_for_non_redirects(self):
        response = Response()
        with self.assertRaises(self.failureException):
            self.assertRedirected(response, to="http://example.com", code=302)

    def test_assertRedirected_assumes_a_302(self):
        response = Response(
            headers=Headers([("Location", ["http://example.com"])]), code=302,
        )
        self.assertRedirected(response, to="http://example.com")
